///@file
/// Base class for canvas placements
//
// Copyright (C) 2012  Thomas Geymayer <tomgey@gmail.com>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA

#include <simgear_config.h>
#include "CanvasPlacement.hxx"
#include <simgear/props/props.hxx>

namespace simgear
{
namespace canvas
{

  //----------------------------------------------------------------------------
  Placement::Placement(SGPropertyNode* node):
    _node(node)
  {

  }

  //----------------------------------------------------------------------------
  Placement::~Placement()
  {

  }

  //----------------------------------------------------------------------------
  SGConstPropertyNode_ptr Placement::getProps() const
  {
    return _node;
  }

  //----------------------------------------------------------------------------
  SGPropertyNode_ptr Placement::getProps()
  {
    return _node;
  }

  //----------------------------------------------------------------------------
  bool Placement::childChanged(SGPropertyNode* child)
  {
    return false;
  }

} // namespace canvas
} // namespace simgear
