// HTTP request keeping response in memory.
//
// Copyright (C) 2013  Thomas Geymayer <tomgey@gmail.com>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA

#include <simgear_config.h>
#include "HTTPMemoryRequest.hxx"

namespace simgear
{
namespace HTTP
{

  //----------------------------------------------------------------------------
  MemoryRequest::MemoryRequest(const std::string& url):
    Request(url, "GET")
  {

  }

  //----------------------------------------------------------------------------
  const std::string& MemoryRequest::responseBody() const
  {
    return _response;
  }

  //----------------------------------------------------------------------------
  void MemoryRequest::responseHeadersComplete()
  {
    Request::responseHeadersComplete();

    if( responseLength() )
      _response.reserve( responseLength() );
  }

  //----------------------------------------------------------------------------
  void MemoryRequest::gotBodyData(const char* s, int n)
  {
    _response.append(s, n);
  }

} // namespace HTTP
} // namespace simgear
