// Copyright (C) 2012 Mathias Froehlich
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//

#ifndef SIMGEAR_READFILECALLBACK_HXX
#define SIMGEAR_READFILECALLBACK_HXX 1

#include <osgDB/Callbacks>

namespace simgear {

/// Read file callback that calls also back into the registrys read file callback.

class OptionsReadFileCallback : public osgDB::ReadFileCallback {
public:
    virtual osgDB::ReaderWriter::ReadResult
    openArchive(const std::string& name, osgDB::ReaderWriter::ArchiveStatus status,
                unsigned int indexBlockSizeHint, const osgDB::Options* options);

    virtual osgDB::ReaderWriter::ReadResult
    readObject(const std::string& name, const osgDB::Options* options);

    virtual osgDB::ReaderWriter::ReadResult
    readImage(const std::string& name, const osgDB::Options* options);
    
    virtual osgDB::ReaderWriter::ReadResult
    readHeightField(const std::string& name, const osgDB::Options* options);

    virtual osgDB::ReaderWriter::ReadResult
    readNode(const std::string& name, const osgDB::Options* options);

    virtual osgDB::ReaderWriter::ReadResult
    readShader(const std::string& name, const osgDB::Options* options);

protected:
    virtual ~OptionsReadFileCallback();
};

}

#endif
