///@file
/// osg::Operation to initialize the OpenVG context used for path rendering
//
// Copyright (C) 2012  Thomas Geymayer <tomgey@gmail.com>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA

#include "VGInitOperation.hxx"
#include <vg/openvg.h>

namespace simgear
{
namespace canvas
{

  //----------------------------------------------------------------------------
  VGInitOperation::VGInitOperation():
    GraphicsOperation("canvas::VGInit", false)
  {

  }


  //----------------------------------------------------------------------------
  void VGInitOperation::operator()(osg::GraphicsContext* context)
  {
    GLint vp[4];
    glGetIntegerv(GL_VIEWPORT, vp);

    // ATTENTION: If using another OpenVG implementation ensure it doesn't
    //            change any OpenGL state!
    vgCreateContextSH(vp[2], vp[3]);
  }

} // namespace canvas
} // namespace simgear
